#############################################################################
##
#W  gapbibl.g             GAP 4 package `browse'                Thomas Breuer
##
#Y  Copyright (C)  2006,  Lehrstuhl D fuer Mathematik,  RWTH Aachen,  Germany
##


#############################################################################
##
#F  BrowseData.HeaderWithRowCounter( <t>, <header>, <nrallrows> )
##
BrowseData.HeaderWithRowCounter:= function( t, header, nrallrows )
    local shown, indir, i;

    shown:= [];
    indir:= t.dynamic.indexRow;
    for i in [ 2, 4 .. Length( t.dynamic.isRejectedRow )-1 ] do
      if not t.dynamic.isRejectedRow[i] then
        AddSet( shown, indir[i] );
      fi;
    od;
    if nrallrows = Length( shown ) then
      return [ "", [ NCurses.attrs.UNDERLINE, true, header,
                     NCurses.attrs.NORMAL,
                     " (", String( nrallrows ), " entries)" ],
               "" ];
    else
      return [ "", [ NCurses.attrs.UNDERLINE, true, header,
                     NCurses.attrs.NORMAL,
                     " (", String( Length( shown ) ), " out of ",
                     String( nrallrows ), " entries)" ],
               "" ];
    fi;
end;


#############################################################################
##
#F  BrowseData.SimplifiedString( <string> )
##
##  Replace non-ASCII unicode characters (via LaTeX encoding),
##  normalize whitespace,
##  replace &LaTeX; notorious characters,
##  insert whitespace behind dots,
##  deal with common umlaut encodings, ...
##
##  It is not obvious how far one wants to go;
##  at least the bibliography of GAP related publications
##  and the bibliography of the main GAP manuals yield acceptable results.
##
##  For source files in BibTeX format,
##  the issue is to replace LaTeX markup by approximations
##  involving only ASCII characters.
##  For source files in GAPDoc's XML based extended format,
##  the issue is to replace non-ASCII unicode characters by approximations
##  involving only ASCII characters.
##  (The latter may become unnecessary in the future.)
##
BrowseData.SimplificationPatterns:= [
  [ "~", " " ],
  [ "\\-", "" ],
  [ "\\accent19 ", "'" ],
  [ "\\accent19", "'" ],
  [ "\\accent127 ", "\"" ],
  [ "\\accent127", "\"" ],
  [ "{\\\"a}", "\"a" ],
  [ "{\\\"o}", "\"o" ],
  [ "{\\\"u}", "\"u" ],
  [ "{\\\"A}", "\"A" ],
  [ "{\\\"O}", "\"O" ],
  [ "{\\\"U}", "\"U" ],
  [ "{\\a'a}", "'a" ],
  [ "{\\a'e}", "'e" ],
  [ "{\\a'o}", "'o" ],
  [ "{\\a'u}", "'u" ],
  [ "{\\a`a}", "`a" ],
  [ "{\\a`e}", "`e" ],
  [ "{\\a`o}", "`o" ],
  [ "{\\a`u}", "`u" ],
  [ "{\\a'A}", "'A" ],
  [ "{\\a'E}", "'E" ],
  [ "{\\a'O}", "'O" ],
  [ "{\\a'U}", "'U" ],
  [ "{\\H{o}}", "\"o" ],
  [ "{\\Ho}", "\"o" ],
  [ "{\\H{u}}", "\"u" ],
  [ "{\\Hu}", "\"u" ],
  [ "{\\textendash}", "--" ],
  [ "{\\nobreakspace}", " " ],
  [ "\\sf ", " " ],
  [ "\\sf{", "{" ],
  [ "\\sf\n", "\n" ],
  [ "{", "" ],
  [ "}", "" ],
  [ "\\", "" ],
  ];

BrowseData.SimplifiedString:= function( string )
    local pair, pos;

    string:= Encode( Unicode( string ), "LaTeXleavemarkup" );
    string:= NormalizedWhitespace( string );
    for pair in BrowseData.SimplificationPatterns do
      string:= ReplacedString( string, pair[1], pair[2] );
    od;
    pos:= Position( string, '.' );
    while pos <> fail do
      if pos < Length( string ) and string[ pos+1 ] <> ' ' then
        string:= Concatenation( string{ [ 1 .. pos ] }, " ",
                                string{ [ pos+1 .. Length( string ) ] } );
      fi;
      pos:= Position( string, '.', pos );
    od;

    return string;
end;


#############################################################################
##
#F  BrowseBibliography( [<bibfiles>] )
##
##  <#GAPDoc Label="Bibliography_section">
##  <Section Label="sec:gapbibl">
##  <Heading>Overview of the &GAP; Bibliography</Heading>
##
##  The &GAP; documentation contains a bibliography of &GAP; related
##  publications, see&nbsp;<Cite Key="GAPBibliography"/>.
##  &Browse; provides access to this information in &GAP;,
##  using the standard facilities of the function
##  <Ref Func="NCurses.BrowseGeneric"/>, i.&nbsp;e.,
##  one can scroll in the list, search for entries, sort by year,
##  sort and categorize by authors etc.
##  <P/>
##  The &Browse; package contains a (perhaps outdated) version of this
##  bibliography.
##  One can get an updated version as follows.
##  <P/>
##  <C>wget -N http://www.gap-system.org/Doc/Bib/gap-publishednicer.bib</C>
##
##  <ManSection>
##  <Func Name="BrowseBibliography" Arg="[bibfiles]"/>
##
##  <Returns>
##  list <C>[ list of bib-strings, list of abbrevs, list of expansions ]</C>.
##  </Returns>
##
##  <Description>
##  This function shows the list of bibliography entries in the files
##  given by the strings in the list <A>bibfiles</A>.
##  If no argument is given then the file <F>bibl/gap-publishednicer.bib</F>
##  in the &Browse; package directory is taken.
##  Another perhaps interesting data file that should be available in the
##  &GAP; distribution is <F>doc/manualbib.xml</F>.
##  This file can be located as follows.
##  <P/>
##  <Example><![CDATA[
##  gap> file:= Filename( DirectoriesLibrary( "doc" ), "manualbib.xml" );;
##  gap> IsExistingFile( file );
##  true
##  ]]></Example>
##  <P/>
##  Besides &BibTeX; format, also the &XML; based extended format provided by
##  the &GAPDoc; package is supported by <Ref Func="BrowseBibliography"/>, see
##  Chapter&nbsp;<Ref Chap="Utilities for Bibliographies" BookName="gapdoc"/>.
##  <P/>
##  In the case of &BibTeX; format input, the problem arises how to show
##  &LaTeX; entities such as accents and umlauts in visual mode.
##  The &ndash;admittedly poor&ndash; function
##  <C>BrowseData.SimplifiedString</C> is used to translate such entities
##  to ASCII text.
##  <P/>
##  In the case of input in the extended format, the problem arises
##  that in visual mode, currently we can show only ASCII characters 
##  (and the symbols in <C>NCurses.lineDraw</C>, but these are handled
##  differently, see Section&nbsp;<Ref Subsect="ssec:ncursesLines"/>).
##  The function <C>BrowseData.SimplifiedString</C> is used to replace
##  other unicode characters by ASCII text.
##  <P/>
##  The return value is a record as returned by
##  <Ref Func="ParseBibXMLExtFiles" BookName="gapdoc"/>,
##  its <C>entries</C> component corresponds to the bibliography entries
##  that have been <Q>clicked</Q> in visual mode.
##  This record can be used as input for
##  <Ref Func="WriteBibFile" BookName="gapdoc"/> or
##  <Ref Func="WriteBibXMLExtFile" BookName="gapdoc"/>,
##  in order to produce a bibliography file,
##  or it can be used as input for
##  <Ref Func="StringBibXMLEntry" BookName="gapdoc"/>,
##  in order to produce strings from the entries, in various formats.
##  <P/>
##  The full functionality of the function
##  <Ref Func="NCurses.BrowseGeneric"/> is available.
##  <P/>
##  <Example><![CDATA[
##  gap> # sort and categorize by year, scroll down, expand a category row
##  gap> BrowseData.SetReplay( "scrrscsedddddxdddddQ" );
##  gap> BrowseBibliography();;
##  gap> # sort & categorize by authors, expand all category rows, scroll down
##  gap> BrowseData.SetReplay( "scscXseddddddQ" );
##  gap> BrowseBibliography();;
##  gap> # sort and categorize by journal, search for a journal name, expand
##  gap> BrowseData.SetReplay( Concatenation( "scrrrsc/J. Algebra",
##  >        [ NCurses.keys.ENTER ], "nxdddQ" ) );
##  gap> BrowseBibliography();;
##  gap> BrowseData.SetReplay( false );
##  ]]></Example>
##  <P/>
##  <E>Implementation remarks</E>:
##  The browse table has a dynamic header (showing the number of entries,
##  which can vary when the table is restricted), no footer and row labels;
##  one row of column labels is given by the descriptions of the table
##  columns (authors, title, year, journal).
##  <P/>
##  Row and column separators are drawn as grids
##  (cf.&nbsp;<Ref Func="NCurses.Grid"/>) composed from the special characters
##  described in Section&nbsp;<Ref Subsect="ssec:ncursesLines"/>,
##  using the component <C>work.SpecialGrid</C> of the browse table,
##  see <Ref Var="BrowseData"/>.
##  <P/>
##  For categorizing by authors, the sort parameter
##  <C>"split rows on categorizing"</C> is set to <C>"yes"</C>,
##  so the authors are distributed to different category rows,
##  hence each entry appears once for each of its authors
##  in the categorized table.
##  When a data row or an entry in a data row is selected,
##  <Q>click</Q> adds the corresponding bibliographhy entry to the result.
##  <P/>
##  The width of the title column is preset; usually titles are too long for
##  one line, and the contents of this column is formatted as a paragraph,
##  using the function <Ref Func="FormatParagraph" BookName="gapdoc"/>.
##  For the authors and journal columns, maximal widths are prescribed,
##  and <Ref Func="FormatParagraph" BookName="gapdoc"/> is used for longer
##  entries.
##  <P/>
##  For three columns, the sort parameters are customized as follows:
##  The <E>authors</E> column does not become hidden when the table is
##  categorized according to this column,
##  sorting by the <E>year</E> yields a <E>de</E>scending order,
##  and the category rows arising from these columns and the <E>journal</E>
##  column show the numbers of the data rows that belong to them.
##  <P/>
##  This function requires some of the utilities provided by the
##  &GAP; package <Package>GAPDoc</Package>
##  (see&nbsp;<Cite Key="GAPDoc"/>),
##  such as <Ref Func="FormatParagraph" BookName="gapdoc"/>,
##  <Ref Func="NormalizeNameAndKey" BookName="gapdoc"/>,
##  <Ref Func="NormalizedNameAndKey" BookName="gapdoc"/>,
##  <Ref Func="ParseBibFiles" BookName="gapdoc"/>,
##  <Ref Func="ParseBibXMLextFiles" BookName="gapdoc"/>,
##  <Ref Func="ParseBibXMLextString" BookName="gapdoc"/>,
##  <Ref Func="RecBibXMLEntry" BookName="gapdoc"/>, and
##  <Ref Func="StringBibAsXMLext" BookName="gapdoc"/>.
##  <P/>
##  The code can be found in the file <F>app/gapbibl.g</F> of the package.
##  </Description>
##  </ManSection>
##  </Section>
##  <#/GAPDoc>
##
BindGlobal( "BrowseBibliography", function( arg )
    local files, header, titlewidth, maxauthorswidth, maxjournalwidth,
          entries, mapping, entities, sources, file, parse, entry, authors,
          title, year, authorswidth, normalize, auth, tit, ye, journal,
          journalwidth, jour, table, ret, result, pos;

    # Get the arguments.
    if   Length( arg ) = 0 then
      files:= [ Filename( DirectoriesPackageLibrary( "browse", "bibl" ),
                          "gap-publishednicer.bib" ) ];
      header:= "GAP Bibliography";
    elif Length( arg ) = 1 and IsString( arg[1] ) then
      files:= arg;
      header:= files[1];
    elif Length( arg ) = 1 and IsList( arg[1] )
                           and not IsEmpty( arg[1] )
                           and ForAll( arg[1], IsString ) then
      files:= arg[1];
      header:= files[1];
      if 1 < Length( files ) then
        header:= Concatenation( header, ", ..." );
      fi;
    else
      Error( "usage: BrowseBibliography( [<bibfiles>] )" );
    fi;

    titlewidth:= 40;
    maxauthorswidth:= 35;
    maxjournalwidth:= 35;

    # Parse the data files.
    entries:= [];
    mapping:= [];
    entities:= [];
    sources:= [];
    for file in files do
      if 3 < Length( file )
         and file{ [ Length( file )-3 .. Length( file ) ] } = ".xml" then
        # Convert the XML structure to records,
        # memorize the source.
        parse:= ParseBibXMLextFiles( file );
        if parse <> fail then
          Append( entries, List( parse.entries,
              x -> RecBibXMLEntry( x, "Text", parse.strings ) ) );
          Append( mapping, parse.strings );
          Append( entities, parse.entities );
          Append( sources, parse.entries );
        fi;
      else
        # Simply convert the BibTeX entries to records,
        # and memorize that no XML format for them was available.
        parse:= ParseBibFiles( file );
        if parse <> fail then
          for entry in parse[1] do
            NormalizeNameAndKey( entry );
          od;
          Append( entries, parse[1] );
          Append( mapping, TransposedMat( parse{ [ 2, 3 ] } ) );
          Append( sources, List( parse[1], x -> fail ) );
        fi;
      fi;
    od;

    # Fetch the items that shall be shown.
    authors:= [];
    title:= [];
    year:= [];
    authorswidth:= 0;
    for entry in entries do

      normalize:= true;
      if   IsBound( entry.authorAsList ) then
        auth:= entry.authorAsList;
      elif IsBound( entry.editorAsList ) then
        auth:= entry.editorAsList;
      elif IsBound( entry.author ) then
        auth:= entry.author;
      elif IsBound( entry.organization ) then
        auth:= entry.organization;
        normalize:= false;
      elif IsBound( entry.editor ) then
        auth:= entry.editor;
      elif IsBound( entry.publisher ) then
        auth:= entry.publisher;
      else
        auth:= "(no author, organization, editor, publisher)";
        normalize:= false;
      fi;
      if normalize and IsString( auth ) then
        # The value comes from a `.bib' file;
        # the value obtained from an `.xml' file need not be changed.
        auth:= NormalizedNameAndKey( auth )[4];
      fi;
      if IsString( auth ) then
        auth:= [ BrowseData.SimplifiedString( auth ) ];
      else
        auth:= List( auth, y -> BrowseData.SimplifiedString(
                                    Concatenation( y[1], ", ", y[2] ) ) );
      fi;
      authorswidth:= Maximum( authorswidth,
                              MaximumList( List( auth, Length ) ) );
      if maxauthorswidth < authorswidth then
        authorswidth:= maxauthorswidth;
        auth:= Concatenation( List( auth, l -> SplitString( FormatParagraph(
          BrowseData.SimplifiedString( l ), maxauthorswidth, "left" ),
          "\n" ) ) );
      fi;
      Add( authors, rec( rows:= auth, align:= "tl" ) );

      if   IsBound( entry.title ) then
        tit:= entry.title;
      elif IsBound( entry.booktitle ) then
        tit:= entry.booktitle;
      else
        tit:= "(no title or booktitle)";
      fi;
      Add( title, rec(
              rows:= SplitString( FormatParagraph(
                       BrowseData.SimplifiedString( tit ),
                                titlewidth, "left" ), "\n" ),
              align:= "tl" ) );

      if IsBound( entry.year ) then
        ye:= entry.year;
      else
        ye:= "(no year)";
      fi;
      Add( year, rec( rows:= [ ye ], align:= "tr" ) );

    od;

    journal:= [];
    journalwidth:= 0;
    for entry in entries do
      if IsBound( entry.journal ) then
        jour:= [ BrowseData.SimplifiedString( entry.journal ) ];
        journalwidth:= Maximum( journalwidth, Length( jour[1] ) );
        if maxjournalwidth < journalwidth then
          journalwidth:= maxjournalwidth;
          jour:= SplitString( FormatParagraph( jour[1], maxjournalwidth,
                                  "left" ), "\n" );
        fi;
        Add( journal, rec( rows:= jour, align:= "tl" ) );
      else
        Add( journal, "" );
      fi;
    od;

    # Construct the browse table.
    table:= rec(
      work:= rec(
        align:= "tl",
        header:= t -> BrowseData.HeaderWithRowCounter( t, header,
                          Length( entries ) ),
        main:= TransposedMat( [ authors, title, year, journal ] ),
        labelsCol:= [ [ rec( rows:= [ "authors" ], align:= "l" ),
                        rec( rows:= [ "title" ], align:= "l" ),
                        rec( rows:= [ "year" ], align:= "r" ),
                        rec( rows:= [ "journal" ], align:= "l" ),
        ] ],
        sepLabelsCol:= "=",
        sepRow:= "-",
        sepCol:= [ "| ", " | ", " | ", " | ", " |" ],

        # Set widths for the authors, title, journal columns.
        widthCol:= [ , authorswidth,, titlewidth,,,, journalwidth ],
        SpecialGrid:= BrowseData.SpecialGridLineDraw,
        Click:= rec(
          select_entry:= rec(
            helplines:= [ "add the BibTeX or XML entry to the result list" ],
            action:= function( t )
            if t.dynamic.selectedEntry <> [ 0, 0 ] then
              Add( t.dynamic.Return,
                   t.dynamic.indexRow[ t.dynamic.selectedEntry[1] ] / 2 );
            fi;
          end ),
          select_row:= ~.work.Click.select_entry,
        ),
      ),
      dynamic:= rec(
        Return:= [],
      ),
    );

    # Customize the sort parameters for the authors, year, journal columns.
    BrowseData.SetSortParameters( table, "column", 1,
        [ "hide on categorizing", "no",
          "add counter on categorizing", "yes",
          "split rows on categorizing", "yes" ] );
    BrowseData.SetSortParameters( table, "column", 3,
        [ "direction", "descending",
          "add counter on categorizing", "yes" ] );
    BrowseData.SetSortParameters( table, "column", 4,
        [ "add counter on categorizing", "yes" ] );

    # Show the table.
    ret:= NCurses.BrowseGeneric( table );

    # Construct the return value.
    result:= rec( entries:= [],
                  strings:= [],
                  entities:= [], );
    if not IsEmpty( ret ) then
      result.strings:= mapping;
      result.entities:= entities;
      for pos in ret do
        if sources[ pos ] = fail then
          ParseBibXMLextString( StringBibAsXMLext( entries[ pos ] ), result );
        else
          Add( result.entries, sources[ pos ] );
        fi;
      od;
    fi;

    return result;
end );


#############################################################################
##
#E

