############################################################################
##
#W nq.gi			NQL				Rene Hartung
##
#H   @(#)$Id: nq.gi,v 1.2 2007/02/28 12:52:21 gap Exp $
##


############################################################################
##
#M  NilpotentQuotientLpGroup ( <LpGroup>, <int> ) . . for invariant LpGroups
##
## Computes the nilpotent quotient G/\gamma_j(G) if a nilpotent quotient
## G/\gamma_i(G) of <LpGroup> has been computed.
##
InstallMethod( NilpotentQuotientLpGroup,
  "for invariantly L-presented groups (with qs) and positive integer",
  true,
  [ IsLpGroup and HasIsInvariantLPresentation and IsInvariantLPresentation
    and HasNilpotentQuotientSystem, IsPosInt], 0,
  function(G,c)
  local Q, 	# current quotient system
        QS,	# old quotient system
	H,	# the nilpotent quotient of <G>
      	i,	# loop variable
        j;	# nilpotency class

  # known quotient system of <G>
  Q:=NilpotentQuotientSystem(G);
 
  # nilpotency class
  j:=Maximum(Q.Weights);
 
  if c=j then 
    # the given nilpotency class <j> is already known
    H:=PcpGroupByCollectorNC(Q.Pccol);
    SetLowerCentralSeriesOfGroup(H,NQL_LCS(Q));

    return(H);
  elif c<j then
    # the given nilpotency class <c> is already computed 
    QS:=SmallerQuotientSystem(Q,c);

    # build the nilpotent quotient with lcs
    H:=PcpGroupByCollectorNC(QS.Pccol);
    SetLowerCentralSeriesOfGroup(H,NQL_LCS(QS));

    return(H);
  else
    if HasLargestNilpotentQuotient(G) then 
      return(LargestNilpotentQuotient(G));
    fi;
    # extend the largest known quotient system
    for i in [j+1..c] do
      QS:=ShallowCopy(Q);

      # extend the quotient system of G/\gamma_i to G/\gamma_{i+1}
      Q:=ExtendQuotientSystem(Q);

      # if we couldn't extend the quotient system any more, we're finished
      if QS.Weights=Q.Weights then 
        SetLargestNilpotentQuotient(G,PcpGroupByCollectorNC(Q.Pccol));
        break;
      fi;

      Info(InfoNQL,1,"Class ",Maximum(Q.Weights),": ",
			 Length(Q.Weights)-Length(QS.Weights),
			 " generators with relative orders: ",
			 RelativeOrders(Q.Pccol)
			   {[Length(QS.Weights)+1..Length(Q.Weights)]});
    od;

    # store the largest nilpotent quotient system
    ResetFilterObj(G,NilpotentQuotientSystem);
    SetNilpotentQuotientSystem(G,Q);

    # build the nilpotent quotient with lcs
    H:=PcpGroupByCollectorNC(Q.Pccol);
    SetLowerCentralSeriesOfGroup(H,NQL_LCS(Q));

    return(H);
  fi;
  end);

############################################################################
##
#M  NilpotentQuotientLpGroup ( <LpGroup>, <int> ) . . for invariant LpGroups 
##
## Computes a nilpotent presentation for G/gamma_i(G)
##
InstallMethod( NilpotentQuotientLpGroup,
  "for invariantly L-presented groups and positive integer",
  true,
  [ IsLpGroup and HasIsInvariantLPresentation and IsInvariantLPresentation,
    IsPosInt ], 0,
  function(G,c)
  local Q,	# current quotient system
	H,	# the nilpotent quotient of <G>
	QS,	# old quotient system
	i;	# loop variable

  # Compute a confluent nilpotent presentation for G/G'
  Q:=InitQuotientSystem(G);

  Info(InfoNQL,1,"Class ",1,": ",Length(Q.Weights),
		     " generators with relative orders: ",
		     RelativeOrders(Q.Pccol));
  
  for i in [1..c-1] do 
    # copy the old quotient system to compare with the extended qs.
    QS:=ShallowCopy(Q);

    # extend the quotient system of G/\gamma_i to G/\gamma_{i+1}
    Q:=ExtendQuotientSystem(Q);
   
    # if we couldn't extend the quotient system any more, we're finished
    if QS.Weights=Q.Weights then 
      SetLargestNilpotentQuotient(G,PcpGroupByCollectorNC(Q.Pccol));
      break;
    fi;

    Info(InfoNQL,1,"Class ",Maximum(Q.Weights),": ",
			 Length(Q.Weights)-Length(QS.Weights),
			 " generators with relative orders: ",
			 RelativeOrders(Q.Pccol)
			   {[Length(QS.Weights)+1..Length(Q.Weights)]});
  od;
  
  # store the largest known nilpotent quotient of <G>
  ResetFilterObj(G,NilpotentQuotientSystem);
  SetNilpotentQuotientSystem(G,Q);

  # build the nilpotent quotient with lcs
  H:=PcpGroupByCollectorNC(Q.Pccol);
  SetLowerCentralSeriesOfGroup(H,NQL_LCS(Q));

  return(H);
  end);

############################################################################
##
#M  NilpotentQuotientLpGroup( <LpGroup> ) . . . . . . for invariant LpGroups
##
## determines the largest nilpotent quotient of an L-presented group
## Attention: this method may not terminate!
##
InstallOtherMethod( NilpotentQuotientLpGroup,
  "for invariantly L-presented groups (with qs)",
  true,
  [ IsLpGroup and HasIsInvariantLPresentation and IsInvariantLPresentation
    and HasNilpotentQuotientSystem ], 0,
  function(G)
  local Q,	# current quotient system
	H,	# largest nilpotent quotient of <G>
	QS,	# old quotient system
	i;	# loop variable

  if HasLargestNilpotentQuotient(G) then 
    return(LargestNilpotentQuotient(G));
  fi;

  # Compute a confluent nilpotent presentation for G/G'
  Q:=NilpotentQuotientSystem(G);
  
  repeat
    QS:=ShallowCopy(Q);
    
    # extend the quotient system of G/\gamma_i to G/\gamma_{i+1}
    Q:=ExtendQuotientSystem(Q);
  
    if not QS.Weights=Q.Weights then 
      Info(InfoNQL,1,"Class ",Maximum(Q.Weights),": ",
			 Length(Q.Weights)-Length(QS.Weights),
			 " generators with relative orders: ",
			 RelativeOrders(Q.Pccol)
			   {[Length(QS.Weights)+1..Length(Q.Weights)]});
    fi;
  until QS.Weights=Q.Weights;
  
  SetLargestNilpotentQuotient(G,PcpGroupByCollectorNC(Q.Pccol));

  # store the largest known nilpotent quotient of <G>
  ResetFilterObj(G,NilpotentQuotientSystem);
  SetNilpotentQuotientSystem(G,Q);

  # build the nilpotent quotient with lcs
  H:=PcpGroupByCollectorNC(Q.Pccol);
  SetLowerCentralSeriesOfGroup(H,NQL_LCS(Q));

  return(H);
  end);

############################################################################
##
#M  NilpotentQuotientLpGroup( <LpGroup> ) . . . . . . for invariant LpGroups
##
## Determines the largest nilpotent quotient of an L-presented group
## Attention: this method may not terminate!
##
InstallOtherMethod( NilpotentQuotientLpGroup,
  "for invariantly L-presented groups",
  true,
  [ IsLpGroup and HasIsInvariantLPresentation and IsInvariantLPresentation ], 0,
  function(G)
  local Q,	# current quotient system
	H,	# largest nilpotent quotient of <G>
	QS,	# old quotient system
	i;	# loop variable

  # Compute a confluent nilpotent presentation for G/G'
  Q:=InitQuotientSystem(G);
  
  repeat
    QS:=ShallowCopy(Q);
    
    # extend the quotient system of G/\gamma_i to G/\gamma_{i+1}
    Q:=ExtendQuotientSystem(Q);
  
    if not QS.Weights=Q.Weights then 
      Info(InfoNQL,1,"Class ",Maximum(Q.Weights),": ",
			 Length(Q.Weights)-Length(QS.Weights),
			 " generators with relative orders: ",
			 RelativeOrders(Q.Pccol)
			   {[Length(QS.Weights)+1..Length(Q.Weights)]});
    fi;
  until QS.Weights=Q.Weights;
  
  SetLargestNilpotentQuotient(G,PcpGroupByCollectorNC(Q.Pccol));

  # store the largest known nilpotent quotient of <G>
  ResetFilterObj(G,NilpotentQuotientSystem);
  SetNilpotentQuotientSystem(G,Q);

  # build the nilpotent quotient with lcs
  H:=PcpGroupByCollectorNC(Q.Pccol);
  SetLowerCentralSeriesOfGroup(H,NQL_LCS(Q));

  return(H);
  end);

############################################################################
##
#M  NqEpimorphismNilpotentQuotientLpGroup ( <LpGroup>, <int> )
##
## Computes an epimorphism from <LpGroup> into its nilpotent quotient 
## G/\gamma_j(G) if a nilpotent quotient quotient system of <LpGroup> 
## has already been computed.
##
InstallMethod( NqEpimorphismNilpotentQuotientLpGroup,
  "for invariantly L-presented groups (with qs.)",
  true,
  [ IsLpGroup and HasIsInvariantLPresentation and IsInvariantLPresentation
    and HasNilpotentQuotientSystem,
    IsPosInt], 0,
  function(G,c)
  local Q, 	# current quotient system
        QS,	# old quotient system
      	i,	# loop variable
        n;	# nilpotency class

  # known quotient system of <G>
  Q:=NilpotentQuotientSystem(G);
 
  # nilpotency class of <Q>
  n:=Maximum(Q.Weights);
 
  if c=n then 
    # the given nilpotency class <n> is already known
    return(Q.Epimorphism);
  elif c<n then
    # the given nilpotency class <c> is already computed 
    QS:=SmallerQuotientSystem(Q,c);
    return(QS.Epimorphism);
  else
    if HasLargestNilpotentQuotient(G) then 
      Info(InfoNQL,1,"Largest nilpotent quotient of class ",
		     NilpotencyClassOfGroup(LargestNilpotentQuotient(G)));
      return(Q.Epimorphism);
    fi;

    # extend the largest known quotient system
    for i in [n+1..c] do
      QS:=ShallowCopy(Q);

      # extend the quotient system of G/\gamma_i to G/\gamma_{i+1}
      Q:=ExtendQuotientSystem(Q);
  
      # if we couldn't extend the quotient system any more, we're finished
      if QS.Weights=Q.Weights then 
        SetLargestNilpotentQuotient(PcpGroupByCollectorNC(Q.Pccol));
        break;
      fi;
    od;

    # store the largest nilpotent quotient system
    ResetFilterObj(G,NilpotentQuotientSystem);
    SetNilpotentQuotientSystem(G,Q);

    return(Q.Epimorphism); 
  fi;
  end);


############################################################################
##
#M  NqEpimorphismNilpotentQuotientLpGroup ( <LpGroup>, <int> )
##
## Computes an epimorphism from <LpGroup> into its nilpotent quotient 
## G/\gamma_<int>(G).
##
InstallMethod( NqEpimorphismNilpotentQuotientLpGroup,
  "for invariantly L-presented groups",
  true,
  [ IsLpGroup and HasIsInvariantLPresentation and IsInvariantLPresentation,
    IsPosInt], 0,
  function(G,c)
  local H; 	# nilpotent quotient <G>/gamma_<c>(<G>)

  # compute the nilpotent quotient of <G>
  H:=NilpotentQuotientLpGroup(G,c);
 
  return(NilpotentQuotientSystem(G).Epimorphism);
  end);

############################################################################
##
#M  NqEpimorphismNilpotentQuotientLpGroup ( <LpGroup> , <PcpGroup> )
##
## Computes an epimorphism from the LpGroup into the nilpotent quotient
## <PcpGroup>.
##
InstallOtherMethod( NqEpimorphismNilpotentQuotientLpGroup,
  "for an L-presented group and its nilpotent quotient",
  true,  
  [ IsLpGroup and HasIsInvariantLPresentation and IsInvariantLPresentation
    and HasNilpotentQuotientSystem, IsPcpGroup ], 0,
  function(G,H)
  local ftl,	# collector of the PcpGroup <H>
	Q,	# largest known nilpotent quotient system of G
     	QS, 	# nilpotent quotient system of <H>
   	imgs,	# images of the epimorphism
        i,	# loop variable
        c,	# nilpotency class of <H>
	n;	# number of generators of G/gamma_c(G)

  # number of generators of the Pcp group (used to determine its qs.)
  ftl:=Collector(H);
  n:=NumberOfGenerators(ftl);

  # the largest known quotient system
  Q:=NilpotentQuotientSystem(G);
  
  # nilpotency class of <H>
  c:=Q.Weights[n];

  # determine the quotient system of <H>
  QS:=SmallerQuotientSystem(Q,c);
 
  imgs:=[];
  for i in [1..Length(QS.Imgs)] do
    if IsInt(QS.Imgs[i]) then 
      imgs[i]:=[QS.Imgs[i],1];
    else
      imgs[i]:=QS.Imgs[i];
    fi;
  od;
  imgs:=List(imgs,x->PcpElementByGenExpList(ftl,x));
  return(GroupHomomorphismByImagesNC(G,H,GeneratorsOfGroup(G),imgs));
  end);

############################################################################
##
#F  NQL_LCS( <QS> )
##
## Computes the lower central series of the nilpotent quotient given by a 
## quotient system <QS>.
##
InstallGlobalFunction( NQL_LCS,
  function(Q)
  local weights,	# weights-function of <Q>
  	i,		# loop variable
  	c, 		# nilpotency class of <Q>
 	H,		# nilpotent presentation corr. to <Q>
  	gens,		# generators of <H>
 	lcs;		# lower central series of <H>

  # nilpotent presentation group corr. to <Q>
  H:=PcpGroupByCollectorNC(Q.Pccol);

  # generators of <H>
  gens:=GeneratorsOfGroup(H);

  # weights function of the given quotient system
  weights:=Q.Weights;

  # nilpotency class of <Q>
  c:=Maximum(weights);

  # build the lower central series
  lcs:=[];
  lcs[c+1]:=SubgroupByIgs(H,[]);
  lcs[1]:=H;
 
  for i in [2..c] do
    # the lower central series as subgroups by an induced generating system
    # with weights at least <i>
    lcs[i]:=SubgroupByIgs(H,gens{Filtered([1..Length(gens)],x->weights[x]>=i)});
  od;

  return(lcs);
  end);


############################################################################
##
#A  LargestNilpotentQuotient( <LpGroup> )
##
InstallMethod( LargestNilpotentQuotient,
  "for L-presented groups",
  true,
  [ IsLpGroup ], 0,
  NilpotentQuotientLpGroup);
